@@ -7,15 +7,9 @@ import traceback
 import ctypes
 import warnings
 import sys
+import os
 
-try:
-    import cairo
-    cairo  # Pyflakes
-    has_cairo = True
-    from gi.repository import Regress as Everything
-except ImportError:
-    has_cairo = False
-
+from gi.repository import Regress as Everything
 from gi.repository import GObject
 from gi.repository import GLib
 from gi.repository import Gio
@@ -26,6 +20,7 @@ try:
 except:
     Gtk = None
 
+from compathelper import PY3
 from helper import capture_exceptions
 
 
@@ -53,7 +48,14 @@ class RawGList(ctypes.Structure):
         return ctypes.POINTER(cls).from_address(id(obj) + offset)
 
 
-@unittest.skipUnless(has_cairo, 'built without cairo support')
+class TestInstanceTransfer(unittest.TestCase):
+
+    def test_main(self):
+        obj = Everything.TestObj()
+        for _ in range(10):
+            obj.instance_method_full()
+
+
 class TestEverything(unittest.TestCase):
 
     def test_bool(self):
@@ -246,7 +248,11 @@ class TestEverything(unittest.TestCase):
         self.assertEqual(Everything.test_utf8_inout(const_str), noconst_str)
 
     def test_filename_return(self):
-        self.assertEqual(Everything.test_filename_return(), ['åäö', '/etc/fstab'])
+        if PY3 and os.name != "nt":
+            result = [os.fsdecode(b'\xc3\xa5\xc3\xa4\xc3\xb6'), '/etc/fstab']
+        else:
+            result = ['åäö', '/etc/fstab']
+        self.assertEqual(Everything.test_filename_return(), result)
 
     def test_int_out_utf8(self):
         # returns g_utf8_strlen() in out argument
@@ -566,10 +572,9 @@ class TestEverything(unittest.TestCase):
             self.assertTrue('TestBoxedPrivate' in str(e_value), str(e_value))
             self.assertTrue('constructor' in str(e_value), str(e_value))
             tb = ''.join(traceback.format_exception(e_type, e_value, e_tb))
-            self.assertTrue('tests/test_everything.py", line' in tb, tb)
+            self.assertTrue('test_everything.py", line' in tb, tb)
 
 
-@unittest.skipUnless(has_cairo, 'built without cairo support')
 class TestNullableArgs(unittest.TestCase):
     def test_in_nullable_hash(self):
         Everything.test_ghash_null_in(None)
@@ -607,7 +612,6 @@ class TestNullableArgs(unittest.TestCase):
         self.assertEqual(None, Everything.TestObj.null_out())
 
 
-@unittest.skipUnless(has_cairo, 'built without cairo support')
 class TestCallbacks(unittest.TestCase):
     called = False
     main_loop = GLib.MainLoop()
@@ -1016,7 +1020,6 @@ class TestCallbacks(unittest.TestCase):
         self.assertEqual(mydict, {'foo': 1, 'bar': 2, 'new': 42})
 
 
-@unittest.skipUnless(has_cairo, 'built without cairo support')
 class TestClosures(unittest.TestCase):
     def test_no_arg(self):
         def callback():
@@ -1077,7 +1080,6 @@ class TestClosures(unittest.TestCase):
         self.assertTrue('return value' in str(exc[0].value), exc[0].value)
 
 
-@unittest.skipUnless(has_cairo, 'built without cairo support')
 class TestBoxed(unittest.TestCase):
     def test_boxed(self):
         object_ = Everything.TestObj()
@@ -1128,7 +1130,6 @@ class TestBoxed(unittest.TestCase):
         self.assertNotEqual(id(boxed), id(copy))
 
 
-@unittest.skipUnless(has_cairo, 'built without cairo support')
 class TestTortureProfile(unittest.TestCase):
     def test_torture_profile(self):
         import time
@@ -1191,7 +1192,6 @@ class TestTortureProfile(unittest.TestCase):
         print("\tTotal: %f sec" % total_time)
 
 
-@unittest.skipUnless(has_cairo, 'built without cairo support')
 class TestAdvancedInterfaces(unittest.TestCase):
     def test_array_objs(self):
         obj1, obj2 = Everything.test_array_fixed_out_objects()
